/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Author: Frdric Loiret
 */

package org.objectweb.fractal.juliac.spoon;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.objectweb.fractal.juliac.CompilationRound;
import org.objectweb.fractal.juliac.CompilationRounds;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.SourceFile;
import org.objectweb.fractal.juliac.api.CompileSupportItf;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.compile.jdt.CompileSupportImpl;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.ucf.UnifiedClassFactoryItf;

import spoon.processing.Environment;
import spoon.reflect.Factory;
import spoon.reflect.JuliacSpoonFactory;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtSimpleType;
import spoon.support.DefaultCoreFactory;
import spoon.support.JavaOutputProcessor;
import spoon.support.StandardEnvironment;
import spoon.support.builder.JDTTreeBuilder;

/**
 * This class implements the services provided by Spoon to Juliac.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Frederic Loiret <Frederic.Loiret@inria.fr>
 * @since 1.3
 */
public class SpoonSupportImpl implements SpoonSupportItf {

    // ----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // ----------------------------------------------------------------------
    
	final public static Class<SpoonSupportItf> SERVICE_TYPE =
		SpoonSupportItf.class;
	
	private Juliac jc;
    private JavaOutputProcessor jop;
    private JuliacSpoonFactory factory;
    private CompileSupportImpl compiler;

    /**
     * Initialize the Spoon service.
     */
    public void init( Juliac jc ) throws IOException {
        
    	this.jc = jc;
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.register(SERVICE_TYPE,this);
        
        /*
         * Spoon relies on the context class loader for loading classes.
         * See for example
         * spoon.support.reflect.reference.CtTypeReferenceImpl#getActualClass().
         * 
         * Although it seems that this is the only location in Spoon 1.3 where
         * getContextClassLoader() is used and that getActualClass() is not used
         * in the Juliac/Spoon code, we set it to prevent bugs in case some
         * future evolutions of the Juliac/Spoon code would use
         * getActualClass().
         */
        ClassLoader cl = jconf.getClassLoader();
        Thread.currentThread().setContextClassLoader(cl);
        
        /*
         * Spoon relies on JDT. Force the initialization of JDT by requesting
         * Juliac to load and instantiate it.
         */
        CompileSupportItf o = jconf.lookupUnique(CompileSupportItf.class);

        if( ! (o instanceof CompileSupportImpl) ) {
            String msg =
                "The Spoon plugin requires the JDT plugin. Got instead: "+o;
            throw new IOException(msg);
        }
        compiler = (CompileSupportImpl) o;
        
        File baseDir = jconf.getBaseDir();
        File genDir = jconf.getGenDir();
        String[] srcs = jconf.getSrcs();
        String[] srclibs = jconf.getSrclibs();
        
        Environment env = new StandardEnvironment();
        jop = new JavaOutputProcessor(genDir);
        env.setDefaultFileGenerator(jop);
        factory = new JuliacSpoonFactory(new DefaultCoreFactory(),env);
        jop.setFactory(factory);
        
        /*
         * Retrieve all source files contained in the source and library
         * locations.
         */
        List<SourceFile> sfs = new ArrayList<SourceFile>();
        for (String src : srcs) {
            SourceFile.addAllJavaFiles(baseDir,src,sfs);
        }        
        for (String srclib : srclibs) {
            SourceFile.addAllJavaFiles(baseDir,srclib,sfs);
        }
        
        /*
         * Build the Spoon code model.
         */
        build(sfs);
        
        /*
         * Register the {@link CtClass} factory.
         */
        UnifiedClassFactoryItf ucf = new UCtClassFactory(jc);
        jc.register(ucf);
    }

    public void close( Juliac jc ) {
    	factory.close();
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.unregister(SERVICE_TYPE,this);
    }
    

    // ----------------------------------------------------------------------
    // Implementation of the SpoonSupportItf interface
    // ----------------------------------------------------------------------
    
    /**
     * Mix the layers whose class names are specified and generate the resulting
     * mixed class.
     * 
     * @param targetClassName  the name of the generated mixed class
     * @param mixins           the list of layer class names
     * @throws IOException     in case of error during the code generation
     */
    public void mixAndGenerate( String targetClassName, List<String> mixins )
    throws IOException {
                
        List<CtClass<?>> layers = SpoonHelper.toCtClass(factory,mixins);
        CtClass<?>[] ctLayers = layers.toArray(new CtClass<?>[layers.size()]);
        MixinClassGenerator mcg = new MixinClassGenerator(factory);
        CtClass<?> generated = mcg.generate(targetClassName,ctLayers);                

        // Dump the generated class
        generateSourceCode(generated);
    }
    
    /**
     * Return the Spoon factory. This method has been added to meet a
     * requirement from Hulotte which provides its own custom Spoon factory.
     * 
     * @since 2.2.2
     * @author Frederic Loiret <Frederic.Loiret@inria.fr>
     */
    @SuppressWarnings("unchecked")
    public Factory getSpoonFactory() {
        return factory;
    }

    /**
     * Build the Spoon code model corresponding to the specified source files.
     */
    public void build( List<SourceFile> sfs ) throws IOException {

        /*
         * Compile the source files.
         */
        List<CompilationUnitDeclaration> unitds = compiler.compile(sfs);

        /*
         * Traverse the compiled code to populate the Spoon code model.
         */
        JDTTreeBuilder tbuilder = new JDTTreeBuilder(factory);
        for (CompilationUnitDeclaration unitd : unitds) {
            unitd.traverse(tbuilder,unitd.scope);
        }
    }
    
    /**
     * Dump the specified element in its corresponding source file.
     * 
     * @throws IllegalArgumentException
     * 		if the specified element does not implement {@link CtSimpleType}
     */
    public void generateSourceCode( Object o )
    throws IOException, IllegalArgumentException {
    	
    	if( !( o instanceof CtSimpleType) ) {
    		String msg = o+" should implement "+CtSimpleType.class.getName();
    		throw new IllegalArgumentException(msg);
    	}
    	CtSimpleType<?> element = (CtSimpleType<?>) o;
    	
        jop.getCreatedFiles().clear();
        jop.process(element);

        String name = element.getQualifiedName();
        List<File> files = jop.getCreatedFiles();
        if( files.size() != 1 ) {
            String msg =
            	"Only one source was expected. Got "+files.size()+" for "+name+
            	".";
            throw new IOException(msg);
        }
        
        JuliacConfig jconf = jc.getJuliacConfig();
        CompilationRounds rounds = jconf.getCompilationRounds();
        CompilationRound round = rounds.getCurrentCompilationRound();
        File genDir = jconf.getGenDir();
        round.addGenerated(SourceFile.create(genDir,name));
    }
}
