/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Jonathan Labejof
 */

package org.objectweb.fractal.juliac.spoon;

import java.lang.annotation.Annotation;
import java.util.List;
import java.util.Set;

import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;

import spoon.reflect.declaration.CtAnnotation;
import spoon.reflect.declaration.CtMethod;
import spoon.reflect.declaration.CtParameter;
import spoon.reflect.declaration.ModifierKind;
import spoon.reflect.reference.CtTypeReference;

/**
 * Implementation of {@link UnifiedMethod} for {@link CtMethod}s.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
 * @since 1.3
 */
public class UCtMethod implements UnifiedMethod {
    
	private Juliac jc;
	
    /** The concrete {@link CtMethod}. */
    private CtMethod<?> ctmethod;
    
    /**
     * The class declaring the current method.
     * 
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     * @since 2.2.5
     */
    private UCtClass uctclass;
    
    public UCtMethod( Juliac jc, UCtClass uctclass, CtMethod<?> ctmethod ) {
    	this.jc = jc;
        this.ctmethod = ctmethod;
        this.uctclass = uctclass;
    }
    
    /**
     * Return the class declaring the current method.
     * 
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     * @since 2.2.5
     */
    public UCtClass getDeclaringClass(){
    	return uctclass;
    }
    
    /**
     * Return the method annotation for the specified annotation type,
     * or <code>null</code> if the method is not annotated.
     */
    public <A extends Annotation> A getAnnotation( Class<A> annotationType ) {
        A a = ctmethod.getAnnotation(annotationType);
        return a;
    }
    
    public Annotation[] getAnnotations() {
    	Set<CtAnnotation<? extends Annotation>> annots =
    		ctmethod.getAnnotations();
    	Annotation[] array = annots.toArray( new Annotation[annots.size()] );
    	return array;
    }
    
    /**
     * Return the types of the exceptions thrown by the current method.
     */
    public UnifiedClass[] getExceptionTypes() {
        Set<CtTypeReference<? extends Throwable>> throwns =
        	ctmethod.getThrownTypes();
        UnifiedClass[] ucs = new UnifiedClass[throwns.size()];
        int i = 0;
        for (CtTypeReference<? extends Throwable> thrown : throwns) {
            ucs[i] = jc.create(thrown.getQualifiedName());
            i++;
        }
        return ucs;
    }

    /**
     * Return the string representations of the generic exception types.
     * 
     * @since 2.1
     */
    public String[] getGenericExceptionTypes() {
        Set<CtTypeReference<? extends Throwable>> throwns =
        	ctmethod.getThrownTypes();
        String[] gets = new String[throwns.size()];
        int i = 0;
        for (CtTypeReference<? extends Throwable> thrown : throwns) {
            gets[i] = SpoonHelper.getClassName(thrown);
            i++;
        }
        return gets;
    }
    
    /**
     * Return a string representation of the generic return type.
     * 
     * @since 2.1
     */
    public String getGenericReturnType() {
        CtTypeReference<?> type = ctmethod.getType();
        String typename = SpoonHelper.getClassName(type);
        return typename;
    }
    
    /**
     * Return the modifiers of the current method.
     */
    public int getModifiers() {
        int result = 0;
        Set<ModifierKind> modks = ctmethod.getModifiers();
        for (ModifierKind modk : modks) {
            int mod = SpoonHelper.toModifier(modk);
            result += mod;
        }
        return result;
    }
    
    /**
     * Return the name of the current method.
     */
    public String getName() {
        return ctmethod.getSimpleName();
    }
    
    /**
     * Return a string representation (the parameter type followed by the
     * space character, followed by the type name) for the current method.
     */
    public String[] getParameters() {
        List<CtParameter<?>> ctparams = ctmethod.getParameters();
        String[] params = new String[ ctparams.size() ];
        int i = 0;
        for (CtParameter<?> ctparam : ctparams) {
            CtTypeReference<?> type = ctparam.getType();
            String typename = SpoonHelper.getClassName(type);

            /*
             * Don't use ctparam.getSimpleName(). The reason is that we want to
             * obtain a result similar to the one obtained with
             * UCMethod.getParameters(). In this cas, and since Java reflection
             * does not provide parameter names, parameters are named arg99.
             */
            StringBuffer sb = new StringBuffer(typename);
            sb.append(" arg");
            sb.append(i);
            params[i] = sb.toString();
            
            i++;
        }
        return params;
    }
    
    /**
     * Return the types of the parameters of the current method.
     */
    public UnifiedClass[] getParameterTypes() {
        List<CtParameter<?>> params = ctmethod.getParameters();
        UnifiedClass[] types = new UnifiedClass[ params.size() ];
        int i = 0;
        for (CtParameter<?> param : params) {
            CtTypeReference<?> tref = param.getType();
            types[i] = jc.create(tref.getQualifiedName());
            i++;
        }
        return types;
    }

    /**
     * Return the return type of the current method.
     */
    public UnifiedClass getReturnType() {
        CtTypeReference<?> tref = ctmethod.getType();
        UnifiedClass uc = jc.create(tref.getQualifiedName());
        return uc;
    }        

    /**
     * Return the names of the type variables declared by the generic
     * declaration for the current method. Return an empty array if the current
     * method does not declare any type variable.
     * 
     * @since 2.2.6
     */
    public String[] getTypeParameterNames() {
        List<CtTypeReference<?>> ftps = ctmethod.getFormalTypeParameters();
        String[] names = SpoonHelper.getFormalTypeParameterNames(ftps);
        return names;
    }
    
    /**
     * Return the string representations of the type variables declared by the
     * generic declaration for the current method. Return an empty array if the
     * current method does not declare any type variable.
     * 
     * @since 2.1
     */
    public String[] getTypeParameters() {
        List<CtTypeReference<?>> ftps = ctmethod.getFormalTypeParameters();
        String[] names = SpoonHelper.getFormalTypeParameters(ftps);
        return names;
    }

    /**
     * Return a string representation of the current method based on the method
     * name and its parameter types such as <code>name(type1,..typen)</code>.
     */
    @Override
    public String toString() {
    	
    	/*
    	 * Do not modify the implementation of this method unless you are sure
    	 * that this does not break
    	 * AbstractProxyClassGenerator#generate(ClassSourceCodeVisitor)
    	 * 
    	 * See the skipping which is implemented for subclasses which redefine
    	 * inherited methods with a more specialized return type.
    	 * 
    	 * This is checked by examples/advanced/foobar-cases/foobar-cases.
    	 */
        
        UnifiedClass[] paramtypes = getParameterTypes();
        
        StringBuffer sb = new StringBuffer(getName());
        sb.append('(');
        for (int i = 0; i < paramtypes.length; i++) {
            UnifiedClass paramtype = paramtypes[i];
            if(i!=0) { sb.append(','); }
            sb.append(paramtype.getName());
        }
        
        sb.append(')');
        return sb.toString();
    }
}
